import copy
import sys

import pyqtgraph as pg
import xarray as xr
from qtpy import QtCore, QtWidgets

from erlab.analysis.fit import MultiPeakModel
from erlab.interactive.utilities import FittingParameterWidget, ParameterGroup

# EDCmultiFitting Igor procedure - 2D wave EDC fit along momentum range

# Momentum range to fit
# Number of bands
# Initial parameters
#   const_bkg
#   shirley_bkg
#   total resolution (fixed?)
#   Temperature (fixed?)
#   Fermi level (fixed?)
#
#   For each band
#       Lor or Gauss
#       Peak intensity
#       Peak position
#       Peak width

LMFIT_METHODS = [
    "leastsq",
    "least_squares",
    "differential_evolution",
    "brute",
    "basinhopping",
    "ampgo",
    "nelder",
    "lbfgsb",
    "powell",
    "cg",
    "newton",
    "cobyla",
    "bfgs",
    "tnc",
    "trust-ncg",
    "trust-exact",
    "trust-krylov",
    "trust-constr",
    "dogleg",
    "slsqp",
    "emcee",
    "shgo",
    "dual_annealing",
]


class SinglePeakWidget(ParameterGroup):
    VALID_LINESHAPE = ["lorentzian", "gaussian"]

    def __init__(self, peak_index):
        self.peak_index = peak_index
        super().__init__(
            **{
                "Peak Shape": {"qwtype": "combobox", "items": self.VALID_LINESHAPE},
                "height": {
                    "qwtype": "fitparam",
                    "showlabel": "Intensity",
                    "name": "height",
                    "spin_kw": {"value": 0.001, "minimumWidth": 100},
                },
                "center": {
                    "qwtype": "fitparam",
                    "showlabel": "Position",
                    "name": "center",
                    "spin_kw": {"value": 0.001, "minimumWidth": 100},
                },
                "width": {
                    "qwtype": "fitparam",
                    "showlabel": "FWHM",
                    "name": "width",
                    "spin_kw": {"value": 0.03, "minimumWidth": 100},
                },
            },
            groupbox_kw={"flat": True, "objectName": f"PeakWidget{self.peak_index}"},
        )

        self.setStyleSheet("QGroupBox#" + self.objectName() + " {border:0;}")
        for w in self.widgets_of_type(FittingParameterWidget):
            w.set_prefix(f"p{int(self.peak_index)}_")

    @property
    def param_dict(self):
        pd = {}
        for w in self.widgets_of_type(FittingParameterWidget):
            pd = pd | w.param_dict
        return pd

    @property
    def peak_shape(self) -> str:
        return self.values["Peak Shape"]


class PlotPeakItem(pg.PlotCurveItem):
    def __init__(self, param_widget: SinglePeakWidget, *args, **kargs):
        self.param_widget = param_widget
        super().__init__(*args, **kargs)
        self._pen_color = self.opts["pen"].color()
        self._pen_width = self.opts["pen"].width()

    def setTempPen(self, *args, **kargs):
        self.opts["pen"] = pg.mkPen(*args, **kargs)
        self.invalidateBounds()
        self.update()

    def setPen(self, *args, **kargs):
        super().setPen(*args, **kargs)
        self._pen_color = self.opts["pen"].color()
        self._pen_width = self.opts["pen"].width()

    def setMouseHover(self, hover):
        # Inform the item that the mouse is (not) hovering over it
        # if self.mouseHovering == hover:
        # return
        # self.mouseHovering = hover
        if hover:
            self.setTempPen(self._pen_color, width=2 * self._pen_width)
        else:
            self.setTempPen(self._pen_color, width=self._pen_width)

        self.update()


class PlotPeakPosition(pg.InfiniteLine):
    def __init__(
        self, param_widget: SinglePeakWidget, curve: PlotPeakItem, *args, **kargs
    ):
        self.param_widget = param_widget
        self.curve = curve
        super().__init__(*args, movable=True, **kargs)

    def mouseDragEvent(self, ev):
        if not self.movable:
            ev.ignore()
            return
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            if ev.isStart():
                self.moving = True
                self.cursorOffset = self.pos() - self.mapToParent(ev.buttonDownPos())
                self.startPosition = self.pos()
                self._start_height = self.param_widget.widgets["height"].value()

            ev.accept()

            if not self.moving:
                return

            pos = self.cursorOffset + self.mapToParent(ev.pos())

            self.param_widget.widgets["center"].setValue(pos.x())
            self.param_widget.widgets["height"].setValue(
                self._start_height + self.mapToParent(ev.pos() - ev.buttonDownPos()).y()
            )

        elif QtCore.Qt.MouseButton.RightButton in ev.buttons():
            if ev.isStart():
                self._start_width = self.param_widget.widgets["width"].value()
            ev.accept()

            val = self.mapToParent(ev.buttonDownPos() - ev.pos()).y()
            y0, _, y1, _ = self.boundingRect().getCoords()
            amount = val / abs(y1 - y0)
            self.param_widget.widgets["width"].setValue(self._start_width + amount)
        if ev.isFinish():
            self.moving = False
            self.sigPositionChangeFinished.emit(self)

        self.setMouseHover(self.moving)

    def refresh_pos(self):
        self.setPos(self.param_widget.widgets["center"].value())

    def setMouseHover(self, hover):
        self.param_widget.parent().parent().setCurrentWidget(self.param_widget)
        self.curve.setMouseHover(hover)
        return super().setMouseHover(hover)


class edctool(QtWidgets.QMainWindow):
    def __init__(self, data, n_bands: int = 1, parameters=None, *args, **kwargs):
        self.data = data

        self.qapp = QtCore.QCoreApplication.instance()
        if not self.qapp:
            self.qapp = QtWidgets.QApplication(sys.argv)
        self.qapp.setStyle("Fusion")
        super().__init__()
        self.resize(720, 360)

        self._dock0 = QtWidgets.QDockWidget("Parameters", self)
        self._options = QtWidgets.QWidget(self)
        self._options_layout = QtWidgets.QVBoxLayout(self._options)
        self._params_init = ParameterGroup(
            n_bands={
                "showlabel": "# Bands",
                "qwtype": "btspin",
                "integer": True,
                "value": n_bands,
                "minimum": 1,
                "fixedWidth": 60,
                "notrack": True,
                "valueChanged": self.refresh_n_peaks,
            },
            lin_bkg={
                "qwtype": "fitparam",
                "showlabel": "Linear Background",
                "name": "lin_bkg",
                "spin_kw": {"value": 0.0, "minimumWidth": 200},
            },
            const_bkg={
                "qwtype": "fitparam",
                "showlabel": "Constant Background",
                "name": "const_bkg",
                "spin_kw": {"value": 0.0, "minimumWidth": 200},
            },
            offset={
                "qwtype": "fitparam",
                "showlabel": "Offset",
                "name": "offset",
                "spin_kw": {"value": 0.0, "minimumWidth": 200},
            },
            resolution={
                "qwtype": "fitparam",
                "showlabel": "Total Resolution",
                "name": "resolution",
                "spin_kw": {
                    "value": 0.01,
                    "singleStep": 0.0001,
                    "decimals": 4,
                    "minimumWidth": 200,
                },
            },
            temp={
                "qwtype": "fitparam",
                "showlabel": "Temperature",
                "name": "temp",
                "fixed": True,
                "spin_kw": {"value": 30, "minimum": 0, "minimumWidth": 200},
            },
            efermi={
                "qwtype": "fitparam",
                "showlabel": "Fermi Level",
                "name": "efermi",
                "fixed": True,
                "spin_kw": {"value": 0, "minimum": 0, "minimumWidth": 200},
            },
            Method={"qwtype": "combobox", "items": LMFIT_METHODS},
            go={
                "qwtype": "pushbtn",
                "showlabel": False,
                "text": "Go",
                "clicked": self.do_fit,
            },
        )
        #

        # label_width = 0
        # for w in self._params_init.widgets_of_type(FittingParameterWidget):
        #     label_width = max(
        #         label_width,
        #         w.label.fontMetrics().boundingRect(w.label.text()).width() + 5,
        #     )
        # for w in self._params_init.widgets_of_type(FittingParameterWidget):
        #     w.label.setFixedWidth(label_width)

        self._options_layout.addWidget(self._params_init)

        self._params_peak = QtWidgets.QTabWidget()

        self._options_layout.addWidget(self._params_peak)

        self._dock0.setWidget(self._options)
        self._dock0.setFloating(False)

        self.plotwidget = pg.PlotWidget(self)
        self.rawplot = self.plotwidget.plot()
        self.rawplot.setData(x=self.xdata, y=self.ydata)

        self.modelplot = self.plotwidget.plot()
        self.modelplot.setPen(pg.mkPen("y"))

        self.fitplot = self.plotwidget.plot()
        self.fitplot.setPen(pg.mkPen("c"))

        self.peakcurves = []
        self.peaklines = []

        self.refresh_n_peaks()

        self._params_init.sigParameterChanged.connect(self._refresh_plot_peaks)

        self.setCentralWidget(self.plotwidget)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, self._dock0)
        self.setWindowTitle("EDC Fitting")

        if parameters is not None:
            self.set_params(parameters)
        self.__post_init__(execute=True)

    @property
    def xdata(self):
        return self.data[self.data.dims[0]].values

    @property
    def ydata(self):
        return self.data.values

    @property
    def n_bands(self):
        return self._params_init.values["n_bands"]

    @property
    def params_dict(self):
        out = {}
        for k in ("efermi", "temp", "lin_bkg", "const_bkg", "offset", "resolution"):
            out = out | self._params_init.widgets[k].param_dict
        for i in range(self.n_bands):
            peak_widget = self._params_peak.widget(i)
            out = out | peak_widget.param_dict
        return out

    @property
    def params(self):
        return {k: v["value"] for k, v in self.params_dict.items()}

    @property
    def model(self):
        return MultiPeakModel(
            self.n_bands,
            peak_shapes=[
                self._params_peak.widget(i).peak_shape for i in range(self.n_bands)
            ],
        )

    def refresh_n_peaks(self):
        current = int(self._params_peak.count())
        if self.n_bands > current:
            while self.n_bands > self._params_peak.count():
                self._params_peak.addTab(SinglePeakWidget(current), f"Peak {current}")
                self._params_peak.widget(current).sigParameterChanged.connect(
                    self._refresh_plot_peaks
                )
                current += 1
        elif self.n_bands == current:
            return
        else:
            while self.n_bands < self._params_peak.count():
                current -= 1
                self._params_peak.removeTab(current)

        self._refresh_plot_peaks()

    def _refresh_plot_peaks(self):
        model = self.model
        params = self.params
        for i in range(self.n_bands):
            peak_widget = self._params_peak.widget(i)
            try:
                curve = self.peakcurves[i]
                line = self.peaklines[i]

            except IndexError:
                curve = PlotPeakItem(peak_widget)
                self.plotwidget.addItem(curve)
                curve.setPen(pg.mkPen("r"))
                self.peakcurves.append(curve)

                line = PlotPeakPosition(peak_widget, curve)
                self.plotwidget.addItem(line)
                line.setPen(pg.mkPen("r"))
                line.setHoverPen(pg.mkPen("r", width=2))
                self.peaklines.append(line)

            curve.setData(x=self.xdata, y=model.func.eval_peak(i, self.xdata, **params))
            line.refresh_pos()

        self.modelplot.setData(x=self.xdata, y=model.eval(x=self.xdata, **params))

    def do_fit(self):
        res = self.model.guess_fit(
            self.data,
            params=self.params_dict,
            method=self._params_init.values["Method"],
        )
        print(res.best_values)
        self.fitplot.setData(x=self.xdata, y=res.best_fit)
        self.set_params(res.best_values)

        self.result = res

    def set_params(self, params: dict):
        params = copy.deepcopy(params)
        self._params_init.set_values(
            **{
                k: params[k]
                for k in (
                    "efermi",
                    "temp",
                    "lin_bkg",
                    "const_bkg",
                    "offset",
                    "resolution",
                )
            }
        )
        for i in range(self.n_bands):
            self._params_peak.widget(i).set_values(
                **{k[3:]: v for k, v in params.items() if k.startswith(f"p{i}")}
            )

    def __post_init__(self, execute=None):
        self.show()
        self.activateWindow()
        # self.raise_()

        if execute is None:
            execute = True
            try:
                shell = get_ipython().__class__.__name__  # type: ignore
                if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                    execute = False
            except NameError:
                pass
        if execute:
            self.qapp.exec()


class mdctool(QtWidgets.QMainWindow):
    def __init__(self, data, n_bands: int = 1, parameters=None, *args, **kwargs):
        self.data = data

        self.qapp = QtCore.QCoreApplication.instance()
        if not self.qapp:
            self.qapp = QtWidgets.QApplication(sys.argv)
        self.qapp.setStyle("Fusion")
        super().__init__()
        self.resize(720, 360)

        self._dock0 = QtWidgets.QDockWidget("Parameters", self)
        self._options = QtWidgets.QWidget(self)
        self._options_layout = QtWidgets.QVBoxLayout(self._options)
        self._params_init = ParameterGroup(
            n_bands={
                "showlabel": "# Bands",
                "qwtype": "btspin",
                "integer": True,
                "value": n_bands,
                "minimum": 1,
                "fixedWidth": 60,
                "notrack": True,
                "valueChanged": self.refresh_n_peaks,
            },
            lin_bkg={
                "qwtype": "fitparam",
                "showlabel": "Linear Background",
                "name": "lin_bkg",
                "spin_kw": {"value": 0.0, "minimumWidth": 200},
            },
            const_bkg={
                "qwtype": "fitparam",
                "showlabel": "Constant Background",
                "name": "const_bkg",
                "spin_kw": {"value": 0.0, "minimumWidth": 200},
            },
            resolution={
                "qwtype": "fitparam",
                "showlabel": "Total Resolution",
                "name": "resolution",
                "spin_kw": {
                    "value": 0.01,
                    "singleStep": 0.0001,
                    "decimals": 4,
                    "minimumWidth": 200,
                },
            },
            Method={"qwtype": "combobox", "items": LMFIT_METHODS},
            go={
                "qwtype": "pushbtn",
                "showlabel": False,
                "text": "Go",
                "clicked": self.do_fit,
            },
        )
        #

        # label_width = 0
        # for w in self._params_init.widgets_of_type(FittingParameterWidget):
        #     label_width = max(
        #         label_width,
        #         w.label.fontMetrics().boundingRect(w.label.text()).width() + 5,
        #     )
        # for w in self._params_init.widgets_of_type(FittingParameterWidget):
        #     w.label.setFixedWidth(label_width)

        self._options_layout.addWidget(self._params_init)

        self._params_peak = QtWidgets.QTabWidget()

        self._options_layout.addWidget(self._params_peak)

        self._dock0.setWidget(self._options)
        self._dock0.setFloating(False)

        self.plotwidget = pg.PlotWidget(self)
        self.rawplot = self.plotwidget.plot()
        self.rawplot.setData(x=self.xdata, y=self.ydata)

        self.modelplot = self.plotwidget.plot()
        self.modelplot.setPen(pg.mkPen("y"))

        self.fitplot = self.plotwidget.plot()
        self.fitplot.setPen(pg.mkPen("c"))

        self.peakcurves = []
        self.peaklines = []

        self.refresh_n_peaks()

        self._params_init.sigParameterChanged.connect(self._refresh_plot_peaks)

        self.setCentralWidget(self.plotwidget)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, self._dock0)
        self.setWindowTitle("MDC Fitting")

        if parameters is not None:
            self.set_params(parameters)
        self.__post_init__(execute=True)

    @property
    def xdata(self):
        return self.data[self.data.dims[0]].values

    @property
    def ydata(self):
        return self.data.values

    @property
    def n_bands(self):
        return self._params_init.values["n_bands"]

    @property
    def params_dict(self):
        out = {}
        for k in ("lin_bkg", "const_bkg", "resolution"):
            out = out | self._params_init.widgets[k].param_dict
        for i in range(self.n_bands):
            peak_widget = self._params_peak.widget(i)
            out = out | peak_widget.param_dict
        return out

    @property
    def params(self):
        return {k: v["value"] for k, v in self.params_dict.items()}

    @property
    def model(self):
        return MultiPeakModel(
            self.n_bands,
            peak_shapes=[
                self._params_peak.widget(i).peak_shape for i in range(self.n_bands)
            ],
            fd=False,
        )

    def refresh_n_peaks(self):
        current = int(self._params_peak.count())
        if self.n_bands > current:
            while self.n_bands > self._params_peak.count():
                self._params_peak.addTab(SinglePeakWidget(current), f"Peak {current}")
                self._params_peak.widget(current).sigParameterChanged.connect(
                    self._refresh_plot_peaks
                )
                current += 1
        elif self.n_bands == current:
            return
        else:
            while self.n_bands < self._params_peak.count():
                current -= 1
                self._params_peak.removeTab(current)

        self._refresh_plot_peaks()

    def _refresh_plot_peaks(self):
        model = self.model
        params = self.params
        for i in range(self.n_bands):
            peak_widget = self._params_peak.widget(i)
            try:
                curve = self.peakcurves[i]
                line = self.peaklines[i]

            except IndexError:
                curve = PlotPeakItem(peak_widget)
                self.plotwidget.addItem(curve)
                curve.setPen(pg.mkPen("r"))
                self.peakcurves.append(curve)

                line = PlotPeakPosition(peak_widget, curve)
                self.plotwidget.addItem(line)
                line.setPen(pg.mkPen("r"))
                line.setHoverPen(pg.mkPen("r", width=2))
                self.peaklines.append(line)

            curve.setData(x=self.xdata, y=model.func.eval_peak(i, self.xdata, **params))
            line.refresh_pos()

        self.modelplot.setData(x=self.xdata, y=model.eval(x=self.xdata, **params))

    def do_fit(self):
        res = self.model.guess_fit(
            self.data,
            params=self.params_dict,
            method=self._params_init.values["Method"],
        )
        print(res.best_values)
        self.fitplot.setData(x=self.xdata, y=res.best_fit)
        self.set_params(res.best_values)

        self.result = res

    def set_params(self, params: dict):
        params = copy.deepcopy(params)
        self._params_init.set_values(
            **{
                k: params[k]
                for k in (
                    "lin_bkg",
                    "const_bkg",
                    "resolution",
                )
            }
        )
        for i in range(self.n_bands):
            self._params_peak.widget(i).set_values(
                **{k[3:]: v for k, v in params.items() if k.startswith(f"p{i}")}
            )

    def __post_init__(self, execute=None):
        self.show()
        self.activateWindow()
        # self.raise_()

        if execute is None:
            execute = True
            try:
                shell = get_ipython().__class__.__name__  # type: ignore
                if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                    execute = False
            except NameError:
                pass
        if execute:
            self.qapp.exec()


if __name__ == "__main__":
    data = xr.open_dataarray(
        "~/Library/CloudStorage/Dropbox-KAIST_12/Kimoon Han/ERLab/Projects/TiSe2 Chiral"
        "/Experiment/220922 ALS BL4/TS2_testedc_2209ALS.nc"
    )
    edctool(
        data,
        3,
        parameters={
            "p0_center": -0.11933563574455268,
            "p0_width": 0.1678003093053242,
            "p0_height": 2041.8150354041834,
            "p1_center": -0.22027597972817484,
            "p1_width": 0.31653868423339476,
            "p1_height": 2750.4023331130743,
            "p2_center": -0.017453580016272105,
            "p2_width": 0.0031789748073378895,
            "p2_height": 7749.551605384519,
            "lin_bkg": 86.67942474925626,
            "const_bkg": 68.66930909421156,
            "efermi": -0.0,
            "temp": 30,
            "offset": 0.7379771332091531,
            "resolution": 0.04826209466467372,
        },
    )



"""Interactive data browser.

.. currentmodule:: erlab.interactive.imagetool

Modules
=======

.. autosummary::
   :toctree:

   core
   slicer
   fastbinning
   controls

"""

from __future__ import annotations

__all__ = ["itool", "ImageTool"]

import gc
import sys
from typing import TYPE_CHECKING

import xarray as xr
from qtpy import QtCore, QtWidgets

import erlab.io
from erlab.interactive.imagetool.controls import (
    ItoolBinningControls,
    ItoolColormapControls,
    ItoolCrosshairControls,
)
from erlab.interactive.imagetool.core import ImageSlicerArea, SlicerLinkProxy
from erlab.interactive.utilities import DictMenuBar, copy_to_clipboard

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np
    import numpy.typing as npt

    from erlab.interactive.imagetool.slicer import ArraySlicer


def itool(
    data: (
        Sequence[xr.DataArray | npt.ArrayLike[np.floating]]
        | xr.DataArray
        | npt.ArrayLike[np.floating]
    ),
    link: bool = False,
    link_colors: bool = True,
    execute: bool | None = None,
    **kwargs,
):
    """Create and display an ImageTool window.

    Parameters
    ----------
    data
        Array-like object or a sequence of such object with 2 to 4 dimensions. See
        notes.
    link
        Whether to enable linking between multiple ImageTool windows, by default
        `False`.
    link_colors
        Whether to link the color maps between multiple linked ImageTool windows, by
        default `True`.
    execute
        Whether to execute the Qt event loop and display the window, by default `None`.
        If `None`, the execution is determined based on the current IPython shell.
    **kwargs
        Additional keyword arguments to be passed onto the underlying slicer area. For a
        full list of supported arguments, see the
        `erlab.interactive.imagetool.core.ImageSlicerArea` documentation.

    Returns
    -------
    ImageTool or tuple of ImageTool
        The created ImageTool window(s).

    Notes
    -----
    - If `data` is a sequence of valid data, multiple ImageTool windows will be created
      and displayed.
    - If `link` is True, the ImageTool windows will be synchronized.

    Examples
    --------
    >>> itool(data, cmap="gray", gamma=0.5)
    >>> itool(data_list, link=True)
    """

    qapp: QtWidgets.QApplication = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)
    qapp.setStyle("Fusion")

    if isinstance(data, list | tuple):
        win = ()
        for d in data:
            win += (ImageTool(d, **kwargs),)
        for w in win:
            w.show()
        win[-1].activateWindow()
        win[-1].raise_()

        if link:
            linker = SlicerLinkProxy(  # noqa: F841
                *[w.slicer_area for w in win], link_colors=link_colors
            )
    else:
        win = ImageTool(data, **kwargs)
        win.show()
        win.raise_()
        win.activateWindow()
    if execute is None:
        execute = True
        try:
            shell = get_ipython().__class__.__name__  # type: ignore
            if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                execute = False
                from IPython.lib.guisupport import start_event_loop_qt4

                start_event_loop_qt4(qapp)
        except NameError:
            pass
    if execute:
        qapp.exec()
        del win
        gc.collect()

        return
    return win


class BaseImageTool(QtWidgets.QMainWindow):
    def __init__(self, data=None, parent=None, **kwargs):
        super().__init__(parent=parent)
        self.slicer_area = ImageSlicerArea(self, data, **kwargs)
        self.setCentralWidget(self.slicer_area)

        self.docks: tuple[QtWidgets.QDockWidget, ...] = tuple(
            QtWidgets.QDockWidget(name, self) for name in ("Cursor", "Color", "Binning")
        )
        for i, d in enumerate(self.docks):
            d.setFeatures(QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetFloatable)
            d.topLevelChanged.connect(
                lambda val, *, idx=i: self._sync_dock_float(val, idx)
            )

        self.docks[0].setWidget(
            self.widget_box(
                ItoolCrosshairControls(
                    self.slicer_area, orientation=QtCore.Qt.Orientation.Vertical
                )
            )
        )
        self.docks[1].setWidget(
            self.widget_box(ItoolColormapControls(self.slicer_area))
        )
        self.docks[2].setWidget(self.widget_box(ItoolBinningControls(self.slicer_area)))

        for d in self.docks:
            self.addDockWidget(QtCore.Qt.DockWidgetArea.TopDockWidgetArea, d)
        self.resize(720, 720)

    @property
    def array_slicer(self) -> ArraySlicer:
        return self.slicer_area.array_slicer

    def _sync_dock_float(self, floating: bool, index: int):
        for i in range(len(self.docks)):
            if i != index:
                self.docks[i].blockSignals(True)
                self.docks[i].setFloating(floating)
                self.docks[i].blockSignals(False)
        self.docks[index].blockSignals(True)
        self.docks[index].setFloating(floating)
        self.docks[index].blockSignals(False)

    # !TODO: this is ugly and temporary, fix it
    def widget_box(self, widget: QtWidgets.QWidget, **kwargs):
        group = QtWidgets.QGroupBox(**kwargs)
        group.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        group_layout = QtWidgets.QVBoxLayout(group)
        group_layout.setContentsMargins(3, 3, 3, 3)
        group_layout.setSpacing(3)
        group_layout.addWidget(widget)
        return group


class ImageTool(BaseImageTool):
    def __init__(self, data=None, **kwargs):
        super().__init__(data, **kwargs)
        self.mnb = ItoolMenuBar(self.slicer_area, self)

        self.slicer_area.sigDataChanged.connect(self.update_title)
        self.update_title()

    def update_title(self):
        if self.slicer_area._data is not None:
            if self.slicer_area._data.name:
                self.setWindowTitle(str(self.slicer_area._data.name))


class ItoolMenuBar(DictMenuBar):
    def __init__(
        self, slicer_area: ImageSlicerArea, parent: QtWidgets.QWidget | None
    ) -> None:
        super().__init__(parent)
        self.slicer_area = slicer_area

        self.createMenus()
        self.refreshMenus()
        self.slicer_area.sigViewOptionChanged.connect(self.refreshMenus)

    @property
    def array_slicer(self) -> ArraySlicer:
        return self.slicer_area.array_slicer

    @property
    def colorAct(self):
        return tuple(
            self.action_dict[k]
            for k in ("colorInvertAct", "highContrastAct", "zeroCenterAct")
        )

    def _generate_menu_kwargs(self) -> dict:
        menu_kwargs = {
            "fileMenu": {
                "title": "&File",
                "actions": {
                    "&Open...": {
                        "shortcut": "Ctrl+O",
                        "triggered": self._open_file,
                    },
                    "&Save As...": {
                        "shortcut": "Ctrl+Shift+S",
                        "triggered": self._export_file,
                    },
                    "&Copy Cursor Values": {
                        "shortcut": "Ctrl+C",
                        "triggered": self._copy_cursor_val,
                    },
                    "&Copy Cursor Indices": {
                        "shortcut": "Ctrl+Alt+C",
                        "triggered": self._copy_cursor_idx,
                    },
                },
            },
            "viewMenu": {
                "title": "&View",
                "actions": {
                    "viewAllAct": {
                        "text": "View &All",
                        "shortcut": "Ctrl+A",
                        "triggered": self.slicer_area.view_all,
                    },
                    "transposeAct": {
                        "text": "&Transpose Main Image",
                        "shortcut": "T",
                        "triggered": lambda: self.slicer_area.swap_axes(0, 1),
                        "sep_after": True,
                    },
                    "addCursorAct": {
                        "text": "&Add New Cursor",
                        "shortcut": "Shift+A",
                        "triggered": self.slicer_area.add_cursor,
                    },
                    "remCursorAct": {
                        "text": "&Remove Current Cursor",
                        "shortcut": "Shift+R",
                        "triggered": self.slicer_area.remove_current_cursor,
                    },
                    "snapCursorAct": {
                        "text": "&Snap to Pixels",
                        "shortcut": "S",
                        "triggered": self.slicer_area.toggle_snap,
                        "checkable": True,
                    },
                    "cursorMoveMenu": {
                        "title": "Cursor Control",
                        "actions": {},
                    },
                    "colorInvertAct": {
                        "text": "Invert",
                        "shortcut": "R",
                        "checkable": True,
                        "toggled": self._set_colormap_options,
                        "sep_before": True,
                    },
                    "highContrastAct": {
                        "text": "High Contrast",
                        "checkable": True,
                        "toggled": self._set_colormap_options,
                    },
                    "zeroCenterAct": {
                        "text": "Center At Zero",
                        "checkable": True,
                        "toggled": self._set_colormap_options,
                        "sep_after": True,
                    },
                },
            },
            "helpMenu": {
                "title": "&Help",
                "actions": {
                    "helpAction": {"text": "DataSlicer Help (WIP)"},
                    "shortcutsAction": {
                        "text": "Keyboard Shortcuts Reference (WIP)",
                        "sep_before": True,
                    },
                },
            },
        }

        menu_kwargs["viewMenu"]["actions"]["cursorMoveMenu"]["actions"][
            "centerCursorAct"
        ] = {
            "text": "&Center Current Cursor",
            "shortcut": "Shift+C",
            "triggered": lambda: self.slicer_area.center_cursor(),
        }
        for i, ((t, s), axis, amount) in enumerate(
            zip(
                (
                    ("Shift Current Cursor Up", "Shift+Up"),
                    ("Shift Current Cursor Down", "Shift+Down"),
                    ("Shift Current Cursor Right", "Shift+Right"),
                    ("Shift Current Cursor Left", "Shift+Left"),
                    ("Shift Current Cursor Up × 10", "Ctrl+Shift+Up"),
                    ("Shift Current Cursor Down × 10", "Ctrl+Shift+Down"),
                    ("Shift Current Cursor Right × 10", "Ctrl+Shift+Right"),
                    ("Shift Current Cursor Left × 10", "Ctrl+Shift+Left"),
                ),
                (1, 1, 0, 0) * 2,
                (1, -1, 1, -1, 10, -10, 10, -10),
            )
        ):
            menu_kwargs["viewMenu"]["actions"]["cursorMoveMenu"]["actions"][
                f"ShiftCursorAct{i}"
            ] = {
                "text": t,
                "shortcut": s,
                "triggered": lambda *,
                ax=self.slicer_area.main_image.display_axis[axis],
                d=amount: self.slicer_area.step_index(ax, d),
            }
        menu_kwargs["viewMenu"]["actions"]["cursorMoveMenu"]["actions"][
            "centerAllCursorsAct"
        ] = {
            "text": "&Center All Cursors",
            "shortcut": "Alt+Shift+C",
            "triggered": lambda: self.slicer_area.center_all_cursors(),
            "sep_before": True,
        }
        for i, ((t, s), axis, amount) in enumerate(
            zip(
                (
                    ("Shift Cursors Up", "Alt+Shift+Up"),
                    ("Shift Cursors Down", "Alt+Shift+Down"),
                    ("Shift Cursors Right", "Alt+Shift+Right"),
                    ("Shift Cursors Left", "Alt+Shift+Left"),
                    ("Shift Cursors Up × 10", "Ctrl+Alt+Shift+Up"),
                    ("Shift Cursors Down × 10", "Ctrl+Alt+Shift+Down"),
                    ("Shift Cursors Right × 10", "Ctrl+Alt+Shift+Right"),
                    ("Shift Cursors Left × 10", "Ctrl+Alt+Shift+Left"),
                ),
                (1, 1, 0, 0) * 2,
                (1, -1, 1, -1, 10, -10, 10, -10),
            )
        ):
            menu_kwargs["viewMenu"]["actions"]["cursorMoveMenu"]["actions"][
                f"ShiftAllCursorAct{i}"
            ] = {
                "text": t,
                "shortcut": s,
                "triggered": lambda *,
                ax=self.slicer_area.main_image.display_axis[axis],
                d=amount: self.slicer_area.step_index_all(ax, d),
            }
        return menu_kwargs

    def createMenus(self):
        menu_kwargs = self._generate_menu_kwargs()
        self.add_items(**menu_kwargs)

        self.menu_dict["viewMenu"].aboutToShow.connect(
            lambda: self.action_dict["remCursorAct"].setDisabled(
                self.slicer_area.n_cursors == 1
            )
        )

    def refreshMenus(self):
        self.action_dict["snapCursorAct"].blockSignals(True)
        self.action_dict["snapCursorAct"].setChecked(self.array_slicer.snap_to_data)
        self.action_dict["snapCursorAct"].blockSignals(False)

        cmap_props = self.slicer_area.colormap_properties
        for ca, k in zip(self.colorAct, ["reversed", "highContrast", "zeroCentered"]):
            ca.blockSignals(True)
            ca.setChecked(cmap_props[k])
            ca.blockSignals(False)

    def _set_colormap_options(self):
        self.slicer_area.set_colormap(
            reversed=self.colorAct[0].isChecked(),
            highContrast=self.colorAct[1].isChecked(),
            zeroCentered=self.colorAct[2].isChecked(),
        )

    def _copy_cursor_val(self):
        copy_to_clipboard(str(self.slicer_area.array_slicer._values))

    def _copy_cursor_idx(self):
        copy_to_clipboard(str(self.slicer_area.array_slicer._indices))

    def _open_file(self):
        valid_files = {
            "xarray HDF5 Files (*.h5)": (xr.load_dataarray, {"engine": "h5netcdf"}),
            "ALS BL4.0.3 Raw Data (*.pxt)": (erlab.io.merlin.load, {}),
            "ALS BL4.0.3 Live (*.ibw)": (erlab.io.merlin.load_live, {}),
            "DA30 Raw Data (*.ibw *.pxt *.zip)": (erlab.io.da30.load, {}),
            "SSRL BL5-2 Raw Data (*.h5)": (erlab.io.ssrl52.load, {}),
            "NetCDF Files (*.nc *.nc4 *.cdf)": (xr.load_dataarray, {}),
        }

        dialog = QtWidgets.QFileDialog(self)
        dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptOpen)
        dialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFile)
        dialog.setNameFilters(valid_files.keys())
        # dialog.setOption(QtWidgets.QFileDialog.Option.DontUseNativeDialog)

        if dialog.exec():
            files = dialog.selectedFiles()
            fn, kargs = valid_files[dialog.selectedNameFilter()]
            # !TODO: handle ambiguous datasets
            self.slicer_area.set_data(fn(files[0], **kargs))
            self.slicer_area.view_all()

    def _export_file(self):
        if self.slicer_area._data is None:
            raise ValueError("Data is Empty!")
        dialog = QtWidgets.QFileDialog(self)
        dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptSave)
        dialog.setFileMode(QtWidgets.QFileDialog.FileMode.AnyFile)
        dialog.setNameFilter(
            "xarray HDF5 Files (*.h5)",
        )
        dialog.setDirectory(f"{self.slicer_area._data.name}.h5")
        # dialog.setOption(QtWidgets.QFileDialog.Option.DontUseNativeDialog)
        if dialog.exec():
            files = dialog.selectedFiles()
            erlab.io.save_as_hdf5(self.slicer_area._data, files[0])


if __name__ == "__main__":
    # import gc
    # import linecache
    # import tracemalloc

    # def display_top(snapshot: tracemalloc.Snapshot, limit=10):
    #     snapshot = snapshot.filter_traces(
    #         (
    #             tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
    #             tracemalloc.Filter(False, "<unknown>"),
    #         )
    #     )
    #     top_stats = snapshot.statistics("traceback")

    #     print(f"\nTop {limit} lines")
    #     for index, stat in enumerate(top_stats[:limit], 1):
    #         frame = stat.traceback[0]
    #         print(
    #             f"#{index} {stat.traceback.total_nframe}: "
    #             f"{frame.filename}:{frame.lineno}: {stat.size/1024:.1f} KiB"
    #         )
    #         line = linecache.getline(frame.filename, frame.lineno).strip()
    #         if line:
    #             print("    %s" % line)

    #     other = top_stats[limit:]
    #     if other:
    #         size = sum(stat.size for stat in other)
    #         print(f"{len(other)} other: {size/1024:.1f} KiB")
    #     total = sum(stat.size for stat in top_stats)
    #     print(f"Total allocated size: {total/1024:.1f} KiB")

    # while True:
    #     try:
    #         idx = int(input("Index: "))
    #         stat = top_stats[idx - 1]
    #     except (IndexError, ValueError):
    #         break
    #     print("%s memory blocks: %.1f KiB" % (stat.count, stat.size / 1024))
    #     for line in stat.traceback.format():
    #         print(line)

    # tracemalloc.start()
    data = xr.load_dataarray(
        #     # "~/Documents/ERLab/TiSe2/kxy10.nc",
        #     # "~/Documents/ERLab/TiSe2/221213_SSRL_BL5-2/fullmap_kconv_.h5",
        #     # "~/Documents/ERLab/CsV3Sb5/2021_Dec_ALS_CV3Sb5/Data/cvs_kxy_small.nc",
        #     "~/Documents/ERLab/CsV3Sb5/2021_Dec_ALS_CV3Sb5/Data/cvs_kxy.nc",
        "~/Documents/ERLab/TiSe2/220410_ALS_BL4/map_mm_4d_.nc",
        # engine="h5netcdf",
    )

    # win = itool(data, bench=True)
    # win = itool(data)
    # del data

    # gc.collect()
    # data = xr.load_dataarray(
    #     "~/Documents/ERLab/CsV3Sb5/2021_Dec_ALS_CV3Sb5/Data/cvs_kxy.nc",
    #     engine="h5netcdf",
    # )

    # from erlab.interactive.exampledata import generate_data

    # data = generate_data()

    # win = itool([data, data], link=True, link_colors=False)
    # win = itool(data.sel(eV=0, method='nearest'))
    win = itool(data)

    # snapshot = tracemalloc.take_snapshot()
    # print(
    #     *[
    #         f"{n} {m * 2**-20:.2f} MB\t"
    #         for n, m in zip(("Current", "Max"), tracemalloc.get_traced_memory())
    #     ],
    #     sep="",
    # )
    # tracemalloc.stop()
    # display_top(snapshot)
    # print(win.array_slicer._nanmeancalls)

    # qapp: QtWidgets.QApplication = QtWidgets.QApplication.instance()
    # if not qapp:
    #     qapp = QtWidgets.QApplication(sys.argv)
    # qapp.setStyle("Fusion")
    # import numpy as np
    # win = ImageTool(np.ones((2,2)))
    # win.show()
    # win.raise_()
    # win.activateWindow()
    # qapp.exec()
